"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
exports.removeColors = exports.LoggingBase = exports.LoggingInterface = exports.DefaultNativeFunctions = void 0;
const consolewriter_js_1 = require("./consolewriter.js");
const inspect_js_1 = require("./inspect.js");
const types_js_1 = require("./types.js");
const browser = typeof window !== "undefined";
exports.DefaultNativeFunctions = {
    startTimer: () => {
        if (browser && window.performance && window.performance.now) {
            return window.performance.now();
        }
        else {
            return Date.now();
        }
    },
    endTimer: (start) => {
        if (browser && window.performance && window.performance.now) {
            return window.performance.now() - start;
        }
        else {
            return Date.now() - start;
        }
    },
};
const consoleAdapter = new consolewriter_js_1.ConsoleAdapter();
const PROJECT_ROOT = typeof process !== "undefined" ? process.cwd() : undefined;
const InitialisedAdapters = Symbol("@hibas123/logging:initialisedAdapters");
class LoggingInterface {
    constructor(names) {
        this.#timerMap = new Map();
        this.#names = names;
        for (const key in this) {
            if (typeof this[key] === "function") {
                this[key] = this[key].bind(this);
            }
        }
    }
    #names;
    #timerMap;
    get names() {
        return [...this.#names];
    }
    debug(...message) {
        this.message(types_js_1.LoggingTypes.Debug, this.#names, message);
    }
    log(...message) {
        this.message(types_js_1.LoggingTypes.Log, this.#names, message);
    }
    warning(...message) {
        this.message(types_js_1.LoggingTypes.Warn, this.#names, message);
    }
    warn(...message) {
        this.message(types_js_1.LoggingTypes.Warn, this.#names, message);
    }
    error(error, ...message) {
        if (!error)
            error = "Empty ERROR was passed, so no informations available";
        if (typeof error === "string") {
            let e = new Error("This is a fake error, to get a stack trace");
            this.message(types_js_1.LoggingTypes.Error, this.#names, [
                error,
                ...message,
                "\n",
                e.stack,
                "\n",
            ]);
        }
        else {
            this.message(types_js_1.LoggingTypes.Error, this.#names, [error.message, "\n", error.stack, "\n", ...message], getCallerFromExisting(error));
        }
    }
    time(id) {
        if (!id)
            id = Math.floor(Math.random() * 899999 + 100000).toString();
        this.#timerMap.set(id, {
            name: id,
            start: LoggingBase.nativeFunctions.startTimer(),
        });
        return {
            id,
            end: () => this.timeEnd(id),
        };
    }
    timeEnd(id) {
        let timer = this.#timerMap.get(id);
        if (timer) {
            let diff = LoggingBase.nativeFunctions.endTimer(timer.start);
            this.message(types_js_1.LoggingTypes.Debug, this.#names, [
                types_js_1.Format.green(`[${timer.name}]`),
                `->`,
                types_js_1.Format.blue(diff.toFixed(4)),
                "ms",
            ]);
            return diff;
        }
        return -1;
    }
}
exports.LoggingInterface = LoggingInterface;
class LoggingBase extends LoggingInterface {
    constructor(options) {
        super(options?.name ? [options.name] : []);
        this.#closed = false;
        this.#logLevel = types_js_1.LoggingTypes.Debug;
        this.#format_map = new types_js_1.DefaultFormatConfig();
        this.#adapters = new Set();
        options = {
            console: true,
            resolve_filename: true,
            ...options,
        };
        if (options.console) {
            this.addAdapter(consoleAdapter);
        }
    }
    #closed;
    #logLevel;
    #resolve_filename;
    #format_map;
    #adapters;
    set logLevel(level) {
        this.#logLevel = level;
    }
    get logLevel() {
        return this.#logLevel;
    }
    async addAdapter(adapter) {
        const init = adapter.init();
        const add = () => {
            this.#adapters.add(adapter);
            if (LoggingBase[InitialisedAdapters].has(adapter)) {
                LoggingBase[InitialisedAdapters].set(adapter, LoggingBase[InitialisedAdapters].get(adapter) + 1);
            }
            else {
                LoggingBase[InitialisedAdapters].set(adapter, 1);
            }
        };
        if (!init) {
            add();
        }
        else {
            Promise.resolve(init).then(add);
        }
    }
    getChild(name) {
        return new LoggingBase.DecoupledLogging([...this.names, name], this);
    }
    message(type, names, message, caller) {
        if (this.#logLevel > type)
            return;
        if (this.#closed) {
            //TODO: Maybe error?
            return;
        }
        const date = new Date();
        const date_str = date.toISOString().replace(/T/, " ").replace(/\..+/, "");
        let file = undefined;
        if (this.#resolve_filename) {
            let file_raw = caller;
            if (!file_raw) {
                try {
                    file_raw = getCallerFile();
                }
                catch (err) {
                    file_raw = {
                        file: "<unknown>",
                        line: 0,
                        column: 0,
                    };
                }
            }
            if (PROJECT_ROOT &&
                file_raw.file &&
                file_raw.file.startsWith(PROJECT_ROOT)) {
                let newF = file_raw.file.substring(PROJECT_ROOT.length);
                if (newF.startsWith("/") || newF.startsWith("\\"))
                    newF = newF.substring(1);
                file_raw.file = newF;
            }
            file = `${file_raw.file || "<unknown>"}:${file_raw.line}:${file_raw.column || 0}`;
        }
        let type_str = types_js_1.LoggingTypes[type].toUpperCase().padEnd(5, " ");
        let type_format;
        switch (type) {
            case types_js_1.LoggingTypes.Log:
                type_format = this.#format_map.log;
                break;
            case types_js_1.LoggingTypes.Error:
                type_format = this.#format_map.error;
                break;
            case types_js_1.LoggingTypes.Debug:
                type_format = this.#format_map.debug;
                break;
            case types_js_1.LoggingTypes.Warn:
                type_format = this.#format_map.warning;
                break;
        }
        const nameFormatted = [];
        if (names.length > 0) {
            nameFormatted.push(new types_js_1.Formatted("["));
            for (let i = 0; i < names.length; i++) {
                nameFormatted.push(new types_js_1.Formatted(names[i], this.#format_map.names));
                if (i < names.length - 1) {
                    nameFormatted.push(this.#format_map.names_delimiter);
                }
            }
            nameFormatted.push(new types_js_1.Formatted("]"));
        }
        let linePrefix = [
            new types_js_1.Formatted("["),
            new types_js_1.Formatted(date_str, this.#format_map.date),
            new types_js_1.Formatted("]["),
            new types_js_1.Formatted(type_str, type_format),
            ...(file
                ? [new types_js_1.Formatted("]["), new types_js_1.Formatted(file, this.#format_map.file)]
                : []),
            new types_js_1.Formatted("]"),
            ...nameFormatted,
            new types_js_1.Formatted(": "),
        ];
        let formattedMessage = [...linePrefix];
        message.forEach((msg, idx) => {
            let format = new types_js_1.Formatted();
            if (msg instanceof types_js_1.Formatted) {
                format = msg;
                msg = msg.content;
            }
            if (typeof msg !== "string") {
                msg = inspect_js_1.default(msg, {
                    colors: false,
                    showHidden: true,
                    depth: 3,
                });
            }
            removeColors(msg)
                .split("\n")
                .forEach((text, index, { length }) => {
                if (index != length - 1) {
                    formattedMessage.push(new types_js_1.Formatted(text + "\n", format), ...linePrefix);
                }
                else {
                    formattedMessage.push(new types_js_1.Formatted(text, format));
                }
            });
            formattedMessage.push(new types_js_1.Formatted(" "));
        });
        let msg = {
            date,
            file,
            names,
            text: formattedMessage,
            type,
        };
        this.#adapters.forEach((adapter) => {
            if (adapter.level <= type) {
                adapter.onMessage(msg);
            }
        });
    }
    async close() {
        if (this.#closed)
            return;
        this.#closed = true;
        for (const adapter of this.#adapters) {
            const cnt = LoggingBase[InitialisedAdapters].get(adapter);
            if (!cnt) {
                //TODO: should not happen!
            }
            else {
                if (cnt <= 1) {
                    if (adapter.close)
                        await adapter.close();
                    LoggingBase[InitialisedAdapters].delete(adapter);
                }
                else {
                    LoggingBase[InitialisedAdapters].set(adapter, cnt - 1);
                }
            }
        }
    }
}
exports.LoggingBase = LoggingBase;
_a = InitialisedAdapters;
LoggingBase[_a] = new Map();
LoggingBase.nativeFunctions = exports.DefaultNativeFunctions;
LoggingBase.DecoupledLogging = class extends LoggingInterface {
    constructor(names, lg) {
        super(names);
        this.#lg = lg;
    }
    #lg;
    message(...params) {
        this.#lg.message(...params);
    }
    getChild(name) {
        return new LoggingBase.DecoupledLogging([...this.names, name], this.#lg);
    }
};
function getStack() {
    // Save original Error.prepareStackTrace
    let origPrepareStackTrace = Error.prepareStackTrace;
    try {
        // Override with function that just returns `stack`
        Error.prepareStackTrace = function (_, stack) {
            return stack;
        };
        // Create a new `Error`, which automatically gets `stack`
        let err = new Error();
        // Evaluate `err.stack`, which calls our new `Error.prepareStackTrace`
        let stack = err.stack;
        return stack;
    }
    finally {
        // Restore original `Error.prepareStackTrace`
        Error.prepareStackTrace = origPrepareStackTrace;
    }
}
function getCallerFile() {
    try {
        let stack = getStack();
        let current_file = stack.shift().getFileName();
        while (stack.length) {
            let caller_file = stack.shift();
            if (current_file !== caller_file.getFileName()) {
                return {
                    file: caller_file.getFileName(),
                    line: caller_file.getLineNumber(),
                    column: caller_file.getColumnNumber(),
                };
            }
        }
    }
    catch (err) { }
    return { file: undefined, line: 0 };
}
function getCallerFromExisting(err) {
    if (!err || !err.stack)
        return { file: "NOFILE", line: 0 };
    let lines = err.stack.split("\n");
    lines.shift(); // removing first line
    while (lines.length > 0) {
        let line = lines.shift();
        let matches = line.match(/[<]?([a-zA-Z]:)?([\/\\]?[a-zA-Z_-])+[.>][a-zA-Z_-]*([:][0-9]+)+/g);
        if (matches && matches.length > 0) {
            let match = matches[0].trim();
            let locationString = match.match(/([:][0-9]+)+$/gm)[0];
            let line;
            let column;
            if (locationString) {
                match = match.slice(0, match.length - locationString.length);
                locationString = locationString.substring(1);
                [line, column] = locationString.split(":").map(Number);
            }
            let file = match;
            return {
                file,
                line,
                column,
            };
        }
    }
}
function removeColors(text) {
    text = text.replace(/[\u001b\u009b][[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-ORZcf-nqry=><]/g, "");
    return text;
}
exports.removeColors = removeColors;
//# sourceMappingURL=base.js.map